/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CGRAPH_FLATTEN_EDGE_HPP
#define CGRAPH_FLATTEN_EDGE_HPP
/**
 * graph flat line means a discretized edge with aequidistand discretization point spaces
 * this class was thought for real graph matching using also slopes between edges for accuracy
 * after trying real graph matching, this seemed to be a NP problem. therefore it's not used anymore
 * but still exists in the source code as dead letters :)
 *
 * this can be used to compare 2 edges for similarities in their derivations
 *
 * the setup is made by 2 steps:
 * 1nd: the existing derivatives of the edge point are copied to the initial zero-flat-edge
 * 2nd: the new derivatives are "flattened" by a gaussian distribution
 *
 * after this 2 setup steps, similar edges can be detected by stretching the shorter edge to
 * the length of the other edge and summing up the differences of the derivative
 */

#include <math.h>
#include <list>


/**
 * cache for gaussian distribution
 */
class CGraphFlattenEdgeGaussian
{
public:
	// create gaussian distribution cache
	float *distr;
	int max_radius;

	/**
	 * call this function first to initialize the gaussian distribution
	 */
	CGraphFlattenEdgeGaussian(int p_max_radius, float sigma)
	{
		float gaussian_const =  1.0f/sqrtf(2.0f*M_PI);		// TODO: implement the tmp*tmp multiplication
		max_radius = p_max_radius;
		distr = new float[max_radius*2+1];

		// dont scale the result -> we use the gaussian distribution as a kind of integral
		for (int x = 0; x < max_radius; x++)
		{
			float tmp = x/sigma;
			float weight =	gaussian_const * expf(-0.5*tmp*tmp);
			distr[max_radius-x] = weight;
			distr[max_radius+x] = weight;
		}
	}

	~CGraphFlattenEdgeGaussian()
	{
		delete distr;
	}
};


/**
 * helper function to flatten a node list 
 */
template <class TNode>
class CGraphFlatLine
{
public:
	class CValue
	{
	public:
		float angle;
		float radius;

		inline CValue(float p_angle, float p_radius)
		{
			angle = p_angle;
			radius = p_radius;
		}

		inline CValue()
		{
			angle = 0;
			radius = 0;
		}

		CValue &operator=(CValue &v)
		{
			angle = v.angle;
			radius = v.radius;
			return *this;
		}
	};

	int size;	// size of graph array
	float resolution;
	CValue *graph;	// discretized graph of the line
	float length;	// length of edge


	/**
	 * constructor for memory allocation
	 */
	CGraphFlatLine(const CGraphFlatLine &f)
	{
		size = f.size;
		resolution = f.resolution;
		length = f.length;
		graph = new CValue[size];

		for (int i = 0; i < size; i++)
			graph[i] = f.graph[i];
	}

	/**
	 * constructor
	 * \param p_resolution	resolution of discretized line
	 * \param angle_power	take the angle to the power of this variable
	 */
	CGraphFlatLine(	list<TNode> &nodes,
				float p_length,
				float p_resolution,
				float angle_power,
				CGraphFlattenEdgeGaussian &gaussian
	)
	{
		resolution = p_resolution;
		length = p_length;

		size = (int)ceilf(length*p_resolution);
		CValue *graphx = new CValue[size];
		graph = new CValue[size];

		// initialize iterators
		typename list<TNode>::iterator iter = nodes.begin();
		if (iter == nodes.end())
			return;

		typename list<TNode>::iterator next_iter = nodes.begin();
		next_iter++;

		if (next_iter == nodes.end())
			return;	// nothing to do

		vec2f prev_vec;	// vector aiming from previous node
		vec2f next_vec;	// vector aiming to next node (computed in every stage)

		TNode &n = *iter;
		TNode &nn = *next_iter;

		prev_vec = nn.pos - n.pos;	// previous vector always aims from previous node to recent node
		float prev_length = prev_vec.length();
		prev_vec = prev_vec.getNormal();

		iter++;
		next_iter++;

		float graph_x = prev_length;	// position in graph
		for (; next_iter != nodes.end(); iter++, next_iter++)
		{
			// load iterators for this and next node
			TNode &n = *iter;
			TNode &nn = *next_iter;

			// next vector always aims from recent node to next node
			next_vec = nn.pos - n.pos;
			float next_length = next_vec.length();
			graph_x += next_length;
			next_vec.normalize();

			// compute angle
			// sin(a) = |v1 x v2| / (|v1|*|v2|)
			float angle = asinf(next_vec.getCrossProd(prev_vec));

			// cos(a) = |v1*v2| / (|v1|*|v2|)
			float dot = next_vec.getDotProd(prev_vec);

			// if the dot product is less than 0, we have to fix the angle
			if (dot < 0)
			{
				if (angle > 0)
					angle = M_PI-angle;
				else
					angle = -M_PI-angle;
			}

			graphx[(int)(graph_x*p_resolution)].angle += angle;


			prev_vec = next_vec;
			prev_length = next_length;
		}

		if (angle_power != 1.0)
		{
			for (int x = 0; x < size; x++)
			{
				float angle = graphx[x].angle;
				if (angle == 0.0)		continue;
				float sign = (angle < 0 ? -1 : 1);
				graphx[x].angle = sign*powf(fabsf(angle)*(1.0f/M_PI), angle_power)*M_PI;
			}
		}

		for (int x = 0; x < size; x++)
		{
			int start_x = x - gaussian.max_radius;
			if (start_x < 0)	start_x = 0;

			int end_x = x + gaussian.max_radius;
			if (end_x >= size)	end_x = size-1;

			float sum_angle = 0;
	//		float sum_count = 0;
			for (int sx = start_x; sx < end_x; sx++)
			{
				sum_angle += graphx[sx].angle*gaussian.distr[gaussian.max_radius + x - sx];
	//			sum_count++;
			}

	//		new_graph[x].angle = sum_angle/sum_count;
			// sum it up without dividing it by count because the angles are discretized
			graph[x].angle = sum_angle;
		}

		delete graphx;
	}


	~CGraphFlatLine()
	{
		delete graph;
	}



	/**
	 * compute the difference between this and another line
	 * \param start0	start index in this line 0 to match flat_edge 1
	 * \param size0		matching size
	 * \param ...1		...
	 */
	float getMatchDifference(
			CGraph<N,T>::CEdgeIterator &iter0,
			float fstart0, float fsize0,

			CGraphFlatLine<TNode> &flat_edge1,
			CGraph<N,T>::CEdgeIterator &iter1,
			float fstart1, float fsize1
	)
	{
		CGraphFlatLine<TNode> *flat_edge[2];
		float fsize[2];
		float fstart[2];

//		std::cout << "-----------" << std::endl;
		/**
		 * length of flat_edge[0] <= length flat_edge[1]
		 * => setup flat_edge array
		 */
		if (fsize0 < fsize1)
		{
			flat_edge[0]	= this;
			fstart[0]	= fstart0;
			fsize[0]	= fsize0;

			flat_edge[1]	= &flat_edge1;
			fstart[1]	= fstart1;
			fsize[1]	= fsize1;
		}
		else
		{
			flat_edge[0]	= &flat_edge1;
			fstart[0]	= fstart1;
			fsize[0]	= fsize1;

			flat_edge[1]	= this;
			fstart[1]	= fstart0;
			fsize[1]	= fsize0;
		}
//		std::cout << "START MD: St." << fstart[0] << " Si." << fsize[0] << " L." << flat_edge[0]->length << "      St." << fstart[1] << " Si." << fsize[1] << " L." << flat_edge[1]->length << std::endl;

		int size[2];
		int start[2];

		if (iter0.reverse)
		{
//			std::cout << "REVERSE: 0" << std::endl;
			// if we work on a reverse operator, start from opposite side
			start[0] = (int)((flat_edge[0]->length - fstart[0])*resolution);
			size[0] = -(int)(fsize[0]*resolution);	// reverse direction
		}
		else
		{
			start[0] = (int)(fstart[0]*resolution);
			size[0] = (int)(fsize[0]*resolution);
		}

		if (iter1.reverse)
		{
//			std::cout << "REVERSE: 1" << std::endl;
			start[1] = (int)((flat_edge[1]->length - fstart[1])*resolution);
			size[1] = -(int)(fsize[1]*resolution);
		}
		else
		{
			start[1] = (int)(fstart[1]*resolution);
			size[1] = (int)(fsize[1]*resolution);
		}

		// abort if size is too small
		if (abs(size[0]) == 0 || abs(size[1]) == 0)
			return 0;	// usually a punish value should be transmitted but because there's no matching distance as well no matching area, return 0

//		std::cout << "0 MD: St." << start[0] << " Si." << size[0] << "    St." << start[1] << " Si." << size[1] << std::endl;

		// check for ARRAY BORDERS for FLAT EDGE 0
		if (start[0] < 0)			start[0] = 0;
		else if (start[0] > flat_edge[0]->size)	start[0] = flat_edge[0]->size-1;

		if (size[0] >= 0)
		{
			if (start[0]+size[0] >= flat_edge[0]->size)
			{
//				std::cout << "SIZE" << size[0];
				size[0] = flat_edge[0]->size - start[0] - 1;
//				std::cout << " to " << size[0] << std::endl;
			}
		}
		else
		{
			if (start[0]+size[0] < 0)
				size[0] = -start[0];
		}

		// check for ARRAY BORDERS for FLAT EDGE 1
		if (start[1] < 0)			start[1] = 0;
		else if (start[1] > flat_edge[1]->size)	start[1] = flat_edge[1]->size-1;

		if (size[1] >= 0)
		{
			if (start[1]+size[1] >= flat_edge[1]->size)
				size[1] = flat_edge[1]->size - start[1] - 1;
		}
		else
		{
			if (start[1]+size[1] < 0)
				size[1] = -start[1];
		}

		return getMatchDifference(	flat_edge[0],	start[0],	size[0],
						flat_edge[1],	start[1],	size[1]
					);
	}

	/**
	 * compute the difference between this and another line
	 * \param start0	start index in this line 0 to match flat_edge 1
	 * \param size0		matching size
	 * 			if size < 0, match in backward direction
	 * 			size0 has to be within the graph array borders!!!
	 * \param ...1		...
	 */
	float getMatchDifference(
			CGraphFlatLine<TNode> *flat_edge0,
			int start0, int size0,

			CGraphFlatLine<TNode> *flat_edge1,
			int start1, int size1
	)
	{
		CGraphFlatLine<TNode> *flat_edge[2];
		int start[2];
		int size[2];
		int abs_size[2];

		/**
		 * length of flat_edge[0] <= length flat_edge[1]
		 * => setup flat_edge array
		 */
		if (abs(size0) < abs(size1))
		{
			flat_edge[0]	= flat_edge0;
			start[0]	= start0;
			size[0]		= size0;
			abs_size[0]	= abs(size0);

			flat_edge[1]	= flat_edge1;
			start[1]	= start1;
			size[1]		= size1;
			abs_size[1]	= abs(size1);
		}
		else
		{
			flat_edge[0]	= flat_edge1;
			start[0]	= start1;
			size[0]		= size1;
			abs_size[0]	= abs(size1);

			flat_edge[1]	= flat_edge0;
			start[1]	= start0;
			size[1]		= size0;
			abs_size[1]	= abs(size0);
		}
//		std::cout << "MD: St." << start[0] << " Si." << size[0] << " ASi." << abs_size[0] << "    St." << start[1] << " Si." << size[1] << " ASi." << abs_size[1] << std::endl;
//		std::cout << "    " << flat_edge[0]->size << " " << flat_edge[1]->size << std::endl;

		/**
		 * compute the difference - stretch the smaller projected line
		 * reference projected line: flat_edge[1]
		 * smaller projected line: flat_edge[0]
		 *
		 * we have to scale the flat_edge0 values because the flat_edge is smaller
		 * if the flat_edge is stretched to the length of flat_edge1, we also have
		 * to decrease the angle values because if there are created more points
		 * with the same angle, the angle is going to be increased
		 */
		float diff_absolute = 0;

		// FLAT_EDGE0 < FLAT_EDGE1!!!

		// if only one size is negative, search backward
		if (size[0]*size[1] >= 0)
		{
			if (size[0] < 0)
			{
				/*
				 * both sides are negative
				 * => start at opposite size
				 *    both iterators are positive
				 */
				start[0] += size[0];
				start[1] += size[1];
			}
//			std::cout << "FORWARD";

			for (int x1 = 0; x1 < abs_size[1]; x1++)
			{
				// recompute p0 position
				int x0 = x1 * abs_size[0] / abs_size[1];	// TODO: maybe we should use floating point here to avoid overflow
				if (x0 >= abs_size[0])	x0 = abs_size[0]-1;
				else if (x0 < 0)	x0 = 0;

				//if (x0 >= flat_edge[0]->size)	std::cout << "         ERROR        " << std::endl;
				//if (x1 >= flat_edge[1]->size)	std::cout << "         ERROR        " << std::endl;

				//if (start[1] + x1 < 0 || start[1] + x1 >= flat_edge[1]->size)	{ std::cout << "    **** ERROR f1 ****   " << std::endl; break; }
				//if (start[0] + x0 < 0 || start[0] + x0 >= flat_edge[0]->size)	{ std::cout << "    **** ERROR f0 ****   " << std::endl; break; }
				diff_absolute += fabsf(flat_edge[1]->graph[start[1] + x1].angle - flat_edge[0]->graph[start[0] + x0].angle);
			}
		}
		else
		{
			/**
			 * one size[x] is negative => switch direction
			 */
			if (size[0] < 0)	start[0] += size[0];
			else 			start[1] += size[1];
//			std::cout << "BACKWARD";

			// backward search
			for (int x1 = 0; x1 < abs_size[1]; x1++)
			{
				int x0 = (x1 * abs_size[0]) / abs_size[1];	// TODO: maybe we should use floating point here to avoid overflow
				if (x0 >= abs_size[0])	x0 = abs_size[0]-1;
				else if (x0 < 0)	x0 = 0;

				//if (x0 >= flat_edge[0]->size)	std::cout << "         ERROR        " << std::endl;
				//if (x1 >= flat_edge[1]->size)	std::cout << "         ERROR        " << std::endl;

				//if (start[1] + x1 < 0 || start[1] + x1 >= flat_edge[1]->size)	{ std::cout << "    **** ERROR b1 ****   " << std::endl; break; }
				//if (start[0] + x0 < 0 || start[0] + x0 >= flat_edge[0]->size)	{ std::cout << "    **** ERROR b0 ****   " << std::endl; break; }
				diff_absolute += fabsf(flat_edge[1]->graph[start[1] + abs_size[1] - x1 - 1].angle + flat_edge[0]->graph[start[0] + x0].angle);
			}
		}
/*
		CGraphFlatLine<TNode> &flat_line0 = *flat_edge[0];
		for (int i = 0; i < flat_line0.size; i++)
			std::cout << flat_line0.graph[i].angle << " ";

		CGraphFlatLine<TNode> &flat_line1 = *flat_edge[0];
		for (int i = 0; i < flat_line1.size; i++)
			std::cout << flat_line1.graph[i].angle << " ";
	
		std::cout << "DIFF ABS: " << diff_absolute << std::endl;
		std::cout << std::endl;
*/
		return diff_absolute;
	}
};

#endif
