/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CGRAPH_SIMPLE_HPP
#define CGRAPH_SIMPLE_HPP
/*
 * simple graph implementation:
 *
 * each node can only have one neighbor
 * the edges are stored as strips and can have more than 2 nodes
 */

#include <list>
#include <limits>
#include "CVector.hh"
#include "CMemoryContainer.hpp"
#include "CGraph.hpp"
#include <iostream>

#define DEBUG_GRAPH	0


/**
 * handler for nodes and graphs
 */
template <int N, typename T>
class CNeighborNode;

/**
 * class to handle a node within the graph
 */
template <int N, typename T>
class CNodeRaw
{
public:
	CVector<N,T> pos;
	float radius;		// radius data
	class CNeighborNode<N,T> *neighbor;	// link to neighbor - 0 if there's no neighbor, we are only able to store one neighbor!

	inline CNodeRaw()
	{
		pos.setZero();
		radius = 0;
		neighbor = NULL;
	}

	inline CNodeRaw &operator=(const CNodeRaw &n)
	{
		pos = n.pos;
		radius = n.radius;
		neighbor = NULL;	// dont copy the neighbor!!!
		return *this;
	}


	/**
	 * return true if the position is the same
	 */
	inline bool operator==(CNodeRaw &n2)
	{
		return pos == n2.pos;
	}
	inline bool operator!=(CNodeRaw &n2)
	{
		return pos != n2.pos;
	}


	inline float &operator[](int i)
	{
		return pos[i];
	}

	inline CNodeRaw(const CNodeRaw &n)
	{
		pos = n.pos;
		radius = n.radius;
		neighbor = NULL;
	}


	inline CNodeRaw(CVector<N,T> &v, T p_radius)
	{
		pos = v;
		radius = p_radius;
		neighbor = NULL;
	}

	inline CNodeRaw(CVector<N,T> v)
	{
		pos = v;
		radius = 0;
		neighbor = NULL;
	}
};

template <typename T>
inline
std::ostream&
operator<<(std::ostream &out, CNodeRaw<2,T> &n)
{
	return out << "[" << n[0] << ", " << n[1] << " | " << n.radius << " | " << (n.neighbor==NULL) << "]";
}

template <typename T>
inline
std::ostream&
operator<<(std::ostream &out, CNodeRaw<3,T> &n)
{
	return out << "[" << n[0] << ", " << n[1] << ", " << n[2] << " | " << n.radius << " | " << (n.neighbor==NULL) << "]";
}

/**
 * neighbor data for neighbor strip
 */
template <int N, typename T>
class CNeighborNode
{
public:
	CNodeRaw<N,T> *node;
	list< CNodeRaw<N,T> > *node_list;
	typename list< CNodeRaw<N,T> >::iterator node_iter;
	typename list< CNodeRaw<N,T> >::reverse_iterator node_rev_iter;

	inline void checkNeighbor()
	{
#if 1
		if (node_list == node->neighbor->node_list)
		{
			// allowed because of circular graphs
//			std::cerr << "the nodes share the same list" << std::endl;
		}

		if (&*node_iter != &*node_rev_iter)
		{
			std::cerr << "iterators dont refer to the same object!" << std::endl;
		}

		if (&*node_iter != node)
		{
			std::cerr << "node_iter doesnt refer to the object!" << std::endl;
		}

		if (node->neighbor == NULL)
		{
			std::cerr << "neighbor node has no neighbor data" << std::endl;
			return;
		}

		CNeighborNode &nb = *(node->neighbor);

		if (&*(nb.node_iter) != &*(nb.node_rev_iter))
		{
			std::cerr << "neighbor: iterators dont refer to the same object!" << std::endl;
		}

		if (&*(nb.node_iter) != nb.node)
		{
			std::cerr << "neighbor: node_iter doesnt refer to the object!" << std::endl;
		}
#endif
	}

	inline CNeighborNode(	list< CNodeRaw<N,T> > &p_node_list,
				typename list< CNodeRaw<N,T> >::iterator &p_iter,
				typename list< CNodeRaw<N,T> >::reverse_iterator &p_rev_iter,
				CNodeRaw<N,T> &p_node
			)
	{
		node = &p_node;
		node_list = &p_node_list;
		node_iter = p_iter;
		node_rev_iter = p_rev_iter;
	}
};

template <int N, typename T>
class CGraphRaw	: public
		list< list< CNodeRaw<N,T> > >
{

public:

	~CGraphRaw()
	{
		cleanup();
	}

	CGraphRaw()
	{
	}

	void cleanup()
	{
		/**
		 * remove pointer data within tree
		 */
		for (typename list< list< CNodeRaw<N,T> > >::iterator gi = this->begin(); gi != this->end(); gi++)
		{
			list< CNodeRaw<N,T> > &nlist = *gi;
			for (typename list< CNodeRaw<N,T> >::iterator li = nlist.begin(); li != nlist.end(); li++)
			{
				CNodeRaw<N,T> &node = *li;
				//if (node.neighbor)
					delete node.neighbor;
			}
		}
	}


	/**
	 * join strips with only 1 neighbors at the endings
	 * thus it joints 2 smaller strips to one larger
	 */
	/*
	CGraphRaw getJoined()
	{
		CGraphRaw joined;

		return joined;
	}
	*/


	/**
	 * store the CGraphRaw to CGraph class
	 *
	 * split strips up into real edges - the separating points are the strip points with neighbors
	 *
	 * after creation of the real edges 2 neighbored edges with |E| == 1 to the adjacent side are joined
	 */
	void toGraph(CGraph<N,T> &graph, bool debugGraph = false)
	{
		graph.cleanup();	// destroy existing edges

		/**
		 * 1st step: insert possible edges
		 *
		 * otherwise we cannot create the links to the neighbored edges
		 */

		for (typename list< list< CNodeRaw<N,T> > >::iterator sli = this->begin(); sli != this->end(); sli++)
		{
			list< CNodeRaw<N,T> > &strip = *sli;	// strip list

			if (strip.size() > 0)
			{
				graph.edges.push_front(typename CGraph<N,T>::CEdge());	// insert first edge
				typename CGraph<N,T>::CEdge *edge = &graph.edges.front();	// load front edge

				typename list< CNodeRaw<N,T> >::iterator si = strip.begin();
				CNodeRaw<N,T> edge_start_node = *si;			// node in list

				CVector<N,T> v(std::numeric_limits<T>::max());
				CNodeRaw<N,T> prev_node = CNodeRaw<N,T>(v, 0);

				for (; si != strip.end(); si++)
				{
					CNodeRaw<N,T> &new_node = *si;	// node in list

					// insert node if it's different from the previous one
					if (new_node != prev_node)
					{
						edge->nodes.push_back(typename CGraph<N,T>::CNode(new_node.pos, new_node.radius));
					}

					if (new_node.neighbor != NULL)
					{
						/*
						 * this node has a neighbor
						 * we have to ceck 2 things before we create an edge
						 */
						// 1st check: node may not be at the beginning of an edge
						if (edge_start_node != new_node)
						{
							// 2nd check: node may not be at the ending of an edge
							// the check is made by going on in the strip and comparing the nodes
							// if the node are not equal, this is not the end of the strip
							typename list< CNodeRaw<N,T> >::iterator tmp_si = si;	// create new iterator
							tmp_si++;	// next node
							for (; tmp_si != strip.end(); tmp_si++)
							{
								if (new_node != *tmp_si)	// if node is not equal to on of the next node (not the ending), break
									break;
							}

							if (tmp_si != strip.end())
							{
								// the strip end was not reached => this is NOT the end
								graph.edges.push_front(typename CGraph<N,T>::CEdge());
								edge = &graph.edges.front();
								edge_start_node = new_node;
								edge->nodes.push_back(typename CGraph<N,T>::CNode(new_node.pos, new_node.radius));
							}
						}
					}

					prev_node = new_node;
				}

				// if there are neighbors on the ending nodes we create a new edge
				// one special case is if there are 2 connected strips by ending nodes because without special handling
				// there can be duplicate parts (extension created by both strips)

				CNodeRaw<N,T> &first_front_node = strip.front();

				// insert edges for the first equal nodes (multiple outgoing edges)
				for (typename list< CNodeRaw<N,T> >::iterator n_si = strip.begin(); n_si != strip.end(); n_si++)
				{
					CNodeRaw<N,T> &node = *n_si;
					if (node.pos != first_front_node.pos)	// abort if there are no more multiple connections at first node
						break;

					if (node.neighbor)
					{
						CNodeRaw<N,T> &nn = *(node.neighbor->node);	// load neighbor node
						if (!graph_link_exists(graph, node, nn))	// check for link between first node and it's neighbor
						{
							graph.edges.push_front(typename CGraph<N,T>::CEdge());
							edge = &graph.edges.front();
							edge->nodes.push_back(typename CGraph<N,T>::CNode(nn.pos, nn.radius));
							edge->nodes.push_back(typename CGraph<N,T>::CNode(node.pos, node.radius));
						}
					}
				}

				CNodeRaw<N,T> &first_back_node = strip.back();

				// insert edges for the first equal nodes (multiple outgoing edges)
				for (typename list< CNodeRaw<N,T> >::reverse_iterator n_si = strip.rbegin(); n_si != strip.rend(); n_si++)
				{
					CNodeRaw<N,T> &node = *n_si;
					if (node.pos != first_back_node.pos)	// abort if there are no more multiple connections at first node
						break;

					if (node.neighbor)
					{
						CNodeRaw<N,T> &nn = *(node.neighbor->node);	// load neighbor node
						if (!graph_link_exists(graph, node, nn))	// check for link between first node and it's neighbor
						{
							graph.edges.push_front(typename CGraph<N,T>::CEdge());
							edge = &graph.edges.front();
							edge->nodes.push_back(typename CGraph<N,T>::CNode(nn.pos, nn.radius));
							edge->nodes.push_back(typename CGraph<N,T>::CNode(node.pos, node.radius));
						}
					}
				}
			}
		}


		/**
		 * 2nd step: initialize the edge neighbors
		 *
		 * the graph is now complete - we have to connect the edges to the neighbored edges
		 * we check every edge endings for similar node positions to other edges
		 */
		int c1 = 0;
		for (typename list< typename CGraph<N,T>::CEdge >::iterator ei1 = graph.edges.begin(); ei1 != graph.edges.end(); ei1++)
		{
			c1++;
//			std::cout << "--- " << c1 << " ---" << std::endl;
			typename CGraph<N,T>::CEdge &edge1 = *ei1;
			typename CGraph<N,T>::CNode &node_front1 = edge1.nodes.front();
			typename CGraph<N,T>::CNode &node_back1 = edge1.nodes.back();

//			std::cout << "FRONT1: " << node_front1 << std::endl;
//			std::cout << "BACK1: " << node_back1 << std::endl;

			int c2 = 0;
			for (typename list< typename CGraph<N,T>::CEdge >::iterator ei2 = graph.edges.begin(); ei2 != graph.edges.end(); ei2++)
			{
				c2++;
				typename CGraph<N,T>::CEdge &edge2 = *ei2;
				if (&edge2 != &edge1)
				{
					typename CGraph<N,T>::CNode &node_front2 = edge2.nodes.front();
					typename CGraph<N,T>::CNode &node_back2 = edge2.nodes.back();
//					std::cout << "FRONT2: " << node_front2 << std::endl;
//					std::cout << "BACK2: " << node_back2 << std::endl;

					if (node_front1 == node_front2)
					{
//						std::cout << "FRONT " << c1 << " <-> FRONT NEIGHBOR " << c2 << std::endl;
						edge1.front_neighbors.push_front(typename CGraph<N,T>::CEdgeNeighbor(&edge2, false));
					}

					if (node_front1 == node_back2)
					{
//						std::cout << "FRONT " << c1 << " <-> BACK NEIGHBOR " << c2 << std::endl;
						edge1.front_neighbors.push_front(typename CGraph<N,T>::CEdgeNeighbor(&edge2, true));
					}
					
					if (node_back1 == node_front2)
					{
//						std::cout << "BACK " << c1 << " <-> FRONT NEIGHBOR " << c2 << std::endl;
						edge1.back_neighbors.push_front(typename CGraph<N,T>::CEdgeNeighbor(&edge2, false));
					}
					
					if (node_back1 == node_back2)
					{
//						std::cout << "BACK " << c1 << " <-> BACK NEIGHBOR " << c2 << std::endl;
						edge1.back_neighbors.push_front(typename CGraph<N,T>::CEdgeNeighbor(&edge2, true));
					}
				}
			}
		}

//std::cout << "A" << std::endl;
		if (debugGraph)
			graph.consistency_check();

//std::cout << "A" << std::endl;

/*
		// DEBUG OUTPUT
		for (typename list<typename CGraph<N,T>::CEdge>::iterator ei1 = graph.edges.begin(); ei1 != graph.edges.end(); ei1++)
		{
			typename CGraph<N,T>::CEdge &edge1 = *ei1;
			CGraph<N,T>::CNode &node_front1 = edge1.nodes.front();
			CGraph<N,T>::CNode &node_back1 = edge1.nodes.back();

			std::cout << "======================== " << edge1.nodes.size() << std::endl;
			std::cout << node_front1 << " " << node_back1 << std::endl;
			std::cout << std::endl;

			for (list<CGraph<N,T>::CNode>::iterator ni = edge1.nodes.begin(); ni != edge1.nodes.end(); ni++)
				std::cout << (*ni).pos << std::endl;
		}
*/
		// 3rd step: join edges connected by a vertex with |E| == 2
		for (typename list<typename CGraph<N,T>::CEdge>::iterator ei = graph.edges.begin(); ei != graph.edges.end(); ei++)
		{
			typename CGraph<N,T>::CEdge &edge = *ei;	// edge: current edge

			if (edge.front_neighbors.size() == 1)	// join if |E| == 2
			{
				// EDGE 1: FRONT CONNECTION

				// load adjacent edge
				typename CGraph<N,T>::CEdgeNeighbor &neighbor = edge.front_neighbors.front();
				typename CGraph<N,T>::CEdge &edge_n = *(neighbor.edge);	// edge_n: neighbor edge

				if (&edge_n == &edge)
				{
					// dont join circled edge (do nothing)
					continue;
				}

				// now we merge edge_n to edge
				if (neighbor.reverse)	// neighbor is connected in reverse iterator order
				{
					// backward append
					typename list<typename CGraph<N,T>::CNode>::reverse_iterator ni = edge_n.nodes.rbegin();
						ni++;	// omit first node (same node)
					for (; ni != edge_n.nodes.rend(); ni++)
						edge.nodes.push_front(typename CGraph<N,T>::CNode((*ni).pos, (*ni).radius));

					// EDGE 2: BACK CONNECTION
					// copy neighbors from back_neighbors to front_neighbors (create connections from joined edge to next edges)
					for (typename list<typename CGraph<N,T>::CEdgeNeighbor>::iterator ni = edge_n.front_neighbors.begin(); ni != edge_n.front_neighbors.end(); ni++)
						edge.front_neighbors.push_back(*ni);
				}
				else
				{
					// forward append
					typename list<typename CGraph<N,T>::CNode>::iterator ni = edge_n.nodes.begin();
						ni++;	// omit first node
					for (; ni != edge_n.nodes.end(); ni++)
						edge.nodes.push_front(typename CGraph<N,T>::CNode((*ni).pos, (*ni).radius));

					// EDGE 2: FORWARD CONNECTION
					// copy neighbors from back_neighbors to front_neighbors (create connections from joined edge to next edges)
					for (typename list<typename CGraph<N,T>::CEdgeNeighbor>::iterator ni = edge_n.back_neighbors.begin(); ni != edge_n.back_neighbors.end(); ni++)
						edge.front_neighbors.push_back(*ni);
				}

				// remove link to neighbor node
				edge.front_neighbors.pop_front();

				// fix the linkage from the other neighbors
				// go to every neighbor
				for (typename list<typename CGraph<N,T>::CEdgeNeighbor>::iterator ni = edge.front_neighbors.begin(); ni != edge.front_neighbors.end(); ni++)
				{
					bool fixed = false;
					typename CGraph<N,T>::CEdge &edge_nn = *(*ni).edge;
//					std::cout << "EDGE_NN: " << &edge_nn << std::endl;

					if ((*ni).reverse)	// reverse lookup -> connection is made with back() node
					{
//						std::cout << "rev" << std::endl;
						for (typename list<typename CGraph<N,T>::CEdgeNeighbor>::iterator nni = edge_nn.back_neighbors.begin(); nni != edge_nn.back_neighbors.end(); nni++)
						{
							if ((*nni).edge == &edge_n)	// "neighbors neighbor" edge is connected to neighbor node of current edge
							{
								(*nni).edge = &edge;
								(*nni).reverse = false;	// connected with edge front nodes
								fixed = true;
								break;
							}
						}
					}
					else
					{
//						std::cout << "nrev" << std::endl;
						for (typename list<typename CGraph<N,T>::CEdgeNeighbor>::iterator nni = edge_nn.front_neighbors.begin(); nni != edge_nn.front_neighbors.end(); nni++)
						{
							if ((*nni).edge == &edge_n)
							{
								(*nni).edge = &edge;
								(*nni).reverse = false;	// connected with edge front nodes
								fixed = true;
								break;
							}
						}
					}

					if (!fixed)
					{
						std::cout << "GRAPH FIX NEIGHBOR LINKS ERROR" << std::endl << std::endl;
						//exit(1);
					}
				}

				// delete edge
				for (typename list<typename CGraph<N,T>::CEdge>::iterator dei = graph.edges.begin(); dei != graph.edges.end(); dei++)
				{
					if (&*dei == &edge_n)
					{
						graph.edges.erase(dei);
						break;
					}
				}

				if (debugGraph)
					graph.consistency_check();
			}

			if (debugGraph)
				graph.consistency_check();

			if (edge.back_neighbors.size() == 1)	// join if |E| == 2
			{
				// EDGE 1: FRONT CONNECTION

				// load adjacent edge
				typename CGraph<N,T>::CEdgeNeighbor &neighbor = edge.back_neighbors.front();
				typename CGraph<N,T>::CEdge &edge_n = *(neighbor.edge);

				if (&edge_n == &edge)
				{
					// dont join circled edge
					continue;
				}

				// now we merge edge_n to edge
				if (neighbor.reverse)	// neighbor is connected in reverse iterator order
				{
					// backward append
					typename list<typename CGraph<N,T>::CNode>::reverse_iterator ni = edge_n.nodes.rbegin();
						ni++;	// omit first node (same node)
					for (; ni != edge_n.nodes.rend(); ni++)
						edge.nodes.push_back(typename CGraph<N,T>::CNode((*ni).pos, (*ni).radius));

					// EDGE 2: BACK CONNECTION
					// copy neighbors from back_neighbors to front_neighbors (create connections from joined edge to next edges)
					for (typename list<typename CGraph<N,T>::CEdgeNeighbor>::iterator ni = edge_n.front_neighbors.begin(); ni != edge_n.front_neighbors.end(); ni++)
						edge.back_neighbors.push_back(*ni);
				}
				else
				{
					// forward append
					typename list<typename CGraph<N,T>::CNode>::iterator ni = edge_n.nodes.begin();
						ni++;	// omit first node
					for (; ni != edge_n.nodes.end(); ni++)
						edge.nodes.push_back(typename CGraph<N,T>::CNode((*ni).pos, (*ni).radius));

					// EDGE 2: FORWARD CONNECTION
					// copy neighbors from back_neighbors to front_neighbors (create connections from joined edge to next edges)
					for (typename list<typename CGraph<N,T>::CEdgeNeighbor>::iterator ni = edge_n.back_neighbors.begin(); ni != edge_n.back_neighbors.end(); ni++)
						edge.back_neighbors.push_back(*ni);
				}

				// remove link to neighbor node
				edge.back_neighbors.pop_front();

				// fix the linkage from the other neighbors
				// visit every neighbor
				for (typename list<typename CGraph<N,T>::CEdgeNeighbor>::iterator ni = edge.back_neighbors.begin(); ni != edge.back_neighbors.end(); ni++)
				{
					bool fixed = false;
					typename CGraph<N,T>::CEdge &edge_nn = *(*ni).edge;

					if ((*ni).reverse)	// reverse lookup -> connection is made with back() node
					{
						for (typename list<typename CGraph<N,T>::CEdgeNeighbor>::iterator nni = edge_nn.back_neighbors.begin(); nni != edge_nn.back_neighbors.end(); nni++)
						{
							if ((*nni).edge == &edge_n)
							{
								(*nni).edge = &edge;
								(*nni).reverse = true;	// connected with edge back nodes
								fixed = true;
								break;
							}
						}
					}
					else
					{
						for (typename list<typename CGraph<N,T>::CEdgeNeighbor>::iterator nni = edge_nn.front_neighbors.begin(); nni != edge_nn.front_neighbors.end(); nni++)
						{
							if ((*nni).edge == &edge_n)
							{
								(*nni).edge = &edge;
								(*nni).reverse = true;	// connected with edge back nodes
								fixed = true;
								break;
							}
						}
					}

					if (!fixed)
					{
						std::cout << "GRAPH FIX NEIGHBOR LINKS ERROR" << std::endl << std::endl;
						exit(1);
					}
				}

				// delete edge
				typename list<typename CGraph<N,T>::CEdge>::iterator dei = graph.edges.begin();
				for (; dei != graph.edges.end(); dei++)
				{
					if (&(*dei) == &edge_n)
					{
						graph.edges.erase(dei);
						break;
					}
				}
				if (dei == graph.edges.end())
					std::cerr << "TO DELETE EDGE NOT FOUND" << std::endl;
			}
			if (debugGraph)
				graph.consistency_check();

			if (edge.back_neighbors.size() == 1)
			{
				// join
			}
		}
/*
		// DEBUG OUTPUT
		for (list<CGraph<N,T>::CEdge>::iterator ei1 = graph.edges.begin(); ei1 != graph.edges.end(); ei1++)
		{
			CGraph<N,T>::CEdge &edge1 = *ei1;
			CGraph<N,T>::CNode &node_front1 = edge1.nodes.front();
			CGraph<N,T>::CNode &node_back1 = edge1.nodes.back();

			std::cout << "======================== " << edge1.nodes.size() << std::endl;
			std::cout << node_front1 << " " << node_back1 << std::endl;
			std::cout << std::endl;

			for (list<CGraph<N,T>::CNode>::iterator ni = edge1.nodes.begin(); ni != edge1.nodes.end(); ni++)
				std::cout << (*ni).pos << std::endl;
		}
*/
		graph.renumEdges();
	}

	bool graph_link_exists(CGraph<N,T> &graph, CNodeRaw<N,T> &node1, CNodeRaw<N,T> &node2)
	{
		for (typename list<typename CGraph<N,T>::CEdge>::iterator ei = graph.edges.begin(); ei != graph.edges.end(); ei++)
		{
			typename CGraph<N,T>::CEdge &edge = *ei;

			typename list<typename CGraph<N,T>::CNode>::iterator node_i = edge.nodes.begin();
			typename list<typename CGraph<N,T>::CNode>::iterator next_i = node_i;
			next_i++;

			for (; next_i != edge.nodes.end(); node_i = next_i, next_i++)
			{
				if (	node1.pos == (*node_i).pos &&
					node2.pos == (*next_i).pos
				)
					return true;

				if (	node2.pos == (*node_i).pos &&
					node1.pos == (*next_i).pos
				)
					return true;
			}
		}
		return false;
	}


	/**
	 * file format:
	 *
	 * int			strip count
	 *
	 * int			nodes in strip #1
	 * float[3]	position and radius for node#1
	 * float[3]	position and radius for node#2
	 * ...
	 * float[3]	position and radius for node#n
	 *
	 * int			nodes in strip #2
	 * float[3]	position and radius for node#1
	 * ...
	 * float[3]	position and radius for node#n
	 *
	 * int		count of neighbordata
	 * int[4]	connect element [1] of strip int[0] with element int[3] of strip int[2]
	 */

	struct SNeighborTmp
	{
		int line1, line2;	// number of line
		int node1, node2;	// number of node
		CNodeRaw<N,T> *node_ptr;		// node1 is set to the node which is already found
		CNodeRaw<N,T> *node_search;	// node_search is set to NULL. everytime a new SNeighborTmp is inserted, first it has to be checked if maybe a counterpart of the inserted node exists
	};

	bool save(CMemoryContainer &container)
	{
		int tmpInt = this->size();
		container.push(tmpInt);

		list<SNeighborTmp> neighborNodeList;

		int lineCount = 0;
		for (typename list< list< CNodeRaw<N,T> > >::iterator gi = this->begin(); gi != this->end(); gi++)
		{
			list< CNodeRaw<N,T> > &node_list = *gi;
			tmpInt = node_list.size();
			container.push(tmpInt);

			int nodeCount = 0;
			for (typename list< CNodeRaw<N,T> >::iterator li = node_list.begin(); li != node_list.end(); li++)
			{
				CNodeRaw<N,T> &n = *li;
				container.push(N, n.pos.data);
				container.push(n.radius);
				if (n.neighbor)
				{
					CNeighborNode<N,T> &nn = *(n.neighbor);
					if (&*(nn.node_iter) != &*(nn.node_rev_iter))
					{
						std::cerr << "FATAL ERROR 1 (" << *(nn.node_iter) << " " << *(nn.node_rev_iter) << " " << ")" << std::endl;
//						container.clear();
//						return false;
					}

					// search for existing neighbor
					typename list< SNeighborTmp >::iterator ni;
					CNodeRaw<N,T> &neighbor_node = *(n.neighbor->node);

					// walk through neighborNodeList to search for existing neighbor link
					for (ni = neighborNodeList.begin(); ni != neighborNodeList.end(); ni++)
					{
						SNeighborTmp &nt = *ni;
						if (nt.node_search != NULL)	// counterpart node already found for this neighbor link
							continue;

						if (nt.node_ptr == &neighbor_node)
						{
							nt.node_search = &n;
							nt.line2 = lineCount;
							nt.node2 = nodeCount;
							break;
						}
					}

					if (ni == neighborNodeList.end())
					{
						// no existing neighbor data was found => insert new neighbor data
						SNeighborTmp newNeighbor = {lineCount, 0, nodeCount, 0, &n, NULL};
						neighborNodeList.push_back(newNeighbor);
					}
				}
				nodeCount++;
			}
			lineCount++;
		}

		std::cout << "  strips (" << (int)this->size() << ")" << std::flush;

		typename list<SNeighborTmp>::iterator ni;
		for (ni = neighborNodeList.begin(); ni != neighborNodeList.end(); ni++)
		{
			SNeighborTmp &nt = *ni;
			if (nt.node_search == NULL)	// counterpart node already found for this neighbor link
			{
				std::cerr << "neighborData is not consistent... aborting" << std::endl;
				return false;
			}
		}

		tmpInt = neighborNodeList.size();
		container.push(tmpInt);

		for (ni = neighborNodeList.begin(); ni != neighborNodeList.end(); ni++)
		{
			SNeighborTmp &n = *ni;
			container.push(n.line1);
			container.push(n.node1);
			container.push(n.line2);
			container.push(n.node2);
		}
		std::cout << "   neighbor connections (" << neighborNodeList.size() << ")" << std::flush;

		return true;
	}


	CNodeRaw<N,T> *getNode(	unsigned int lineNr,
			unsigned int nodeNr,
			list< CNodeRaw<N,T> > **list1,
			typename list< CNodeRaw<N,T> >::iterator &iter1,
			typename list< CNodeRaw<N,T> >::reverse_iterator &rev_iter1
		)
	{
//		std::cout << "N: " << lineNr << " " << nodeNr << std::endl;
		typename list< list< CNodeRaw<N,T> > >::iterator gi = this->begin();

//		std::cout << " count count: " << this->size() << std::endl;
		for (unsigned int lineCount = 0; lineCount < lineNr; lineCount++)
		{
			if (gi == this->end())
			{
				std::cerr << "unexpected end of graph lines... aborting" << std::endl;
				return NULL;
			}
			gi++;
		}
		list< CNodeRaw<N,T> > &nlist = *gi;
		*list1 = &nlist;

		iter1 = (*gi).begin();
		for (unsigned int nodeCount = 0; nodeCount < nodeNr; nodeCount++)
		{
//			std::cout << nodeCount << " i " << (*li).pos << std::endl;
			iter1++;
			if (iter1 == (*gi).end())
			{
				std::cerr << "unexpected end of line (forward iterating)... aborting" << std::endl;
				return NULL;
			}
		}

		typename list< CNodeRaw<N,T> >::reverse_iterator rli;
		for (rev_iter1 = nlist.rbegin(); &*rev_iter1 != &*iter1; rev_iter1++);

		return &(*iter1);
	}



	bool load(CMemoryContainer &container)
	{
		this->cleanup();
		this->clear();
		int tmpInt;

		int listSize;
		container.get(&listSize);

		for (int lineCount = 0; lineCount < listSize; lineCount++)
		{
			push_back(list< CNodeRaw<N,T> >());
			list< CNodeRaw<N,T> > &node_list = this->back();

			container.get(&tmpInt);
			int nodeSize = tmpInt;

			for (int nodeCount = 0; nodeCount < nodeSize; nodeCount++)
			{
				node_list.push_back(CNodeRaw<N,T>());

				CNodeRaw<N,T> &node = node_list.back();
				container.get(N, node.pos.data);
				container.get(&(node.radius));
			}
		}
		std::cout << "  strips (" << listSize << ")" << std::flush;

		int neighborSize;
		container.get(&neighborSize);

		for (int neighborCount = 0; neighborCount < neighborSize; neighborCount++)
		{
			int lineNr, nodeNr;
			container.get(&lineNr);
			container.get(&nodeNr);
			list< CNodeRaw<N,T> > *list1 = NULL;
			typename list< CNodeRaw<N,T> >::iterator iter1;
			typename list< CNodeRaw<N,T> >::reverse_iterator rev_iter1;
			CNodeRaw<N,T> *n1 = getNode(lineNr, nodeNr, &list1, iter1, rev_iter1);

			container.get(&lineNr);
			container.get(&nodeNr);
			list< CNodeRaw<N,T> > *list2 = NULL;
			typename list< CNodeRaw<N,T> >::iterator iter2;
			typename list< CNodeRaw<N,T> >::reverse_iterator rev_iter2;
			CNodeRaw<N,T> *n2 = getNode(lineNr, nodeNr, &list2, iter2, rev_iter2);

			if (n1 == NULL || n2 == NULL)
			{
				std::cerr << "cannot find node for cross connection... aborting" << std::endl;
				return false;
			}

			n1->neighbor = new CNeighborNode<N,T>(*list2, iter2, rev_iter2, *n2);
			n2->neighbor = new CNeighborNode<N,T>(*list1, iter1, rev_iter1, *n1);
		}
		std::cout << "   neighbor connections (" << neighborSize << ")" << std::flush;

		return true;
	}
};




#endif
