/*
 * This file is part of the segmentation and registration toolkit (SphereSegRegTk).
 *
 * SphereSegRegTk is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * SphereSegRegTk is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SphereSegRegTk.  If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * math helper functions
 */
#ifndef LBMATH_H
#define LBMATH_H

#include <iostream>
#include <math.h>
#include <stdlib.h>


#ifndef fabsf
#define	m_fabsf(x)	(::fabs(x))
#else
#define	m_fabsf(x)	(::fabsf(x))
#endif

#ifndef floorf
#define	m_floorf(x)	(::floor(x))
#else
#define	m_floorf(x)	(::floorf(x))
#endif

#ifndef ceilf
#define	m_ceilf(x)	(::ceil(x))
#else
#define	m_ceilf(x)	(::ceilf(x))
#endif

#ifndef sqrtf
#define	m_sqrtf(x)	(::sqrt(x))
#else
#define	m_sqrtf(x)	(::sqrtf(x))
#endif

#ifndef sqrtl
#define	m_sqrtl(x)	((unsigned long)::sqrt((double)x))
#else
#define	m_sqrtl(x)	((unsigned long)::sqrtl((unsigned long)x))
#endif


#ifndef roundf
#define m_roundf(x)	((float)(int)(x+0.5f))
#else
#define m_roundf(x)	(::roundf(x))
#endif

#ifndef round
#define m_round(x)	((double)(int)(x+0.5))
#else
#define m_round(x)	(round(x))
#endif


template <typename T>
class CMath
{
public:

	inline CMath()	{
	}

	inline CMath(T a)	{	value = a;	}

// CMath numbers
	inline CMath	operator+(const CMath &a)	{	return CMath(value + a.value);	}
	inline CMath	operator-(const CMath &a)	{	return CMath(value - a.value);	}
	inline CMath	operator*(const CMath a)	{	return CMath(value * a.value);	}
	inline CMath	operator/(const CMath &a)	{	return CMath(value / a.value);	}

	inline CMath&	operator+=(const CMath &a)	{	value += a.value;		return *this;	}
	inline CMath&	operator-=(const CMath &a)	{	value -= a.value;		return *this;	}
	inline CMath&	operator*=(const CMath &a)	{	value *= a.value;		return *this;	}
	inline CMath&	operator/=(const CMath &a)	{	value /= a.value;		return *this;	}

// T numbers
	inline CMath	operator+(const T a)		{	return CMath(value + a);	}
	inline CMath	operator-(const T a)		{	return CMath(value - a);	}
	inline CMath	operator*(const T a)		{	return CMath(value * a);	}
	inline CMath	operator/(const T a)		{	return CMath(value / a);	}
	inline CMath&	operator+=(const T a)		{	value += a;		return *this;	}
	inline CMath&	operator-=(const T a)		{	value -= a;		return *this;	}
	inline CMath&	operator*=(const T a)		{	value *= a;		return *this;	}
	inline CMath&	operator/=(const T a)		{	value /= a;		return *this;	}

	static inline T abs(T a);
	static inline T floor(T a);
	static inline T ceil(T a);
	static inline T sqrt(T a);
	static inline T round(T a);

	static inline T max(T a, T b)		{	return (a < b ? b : a);	}
	static inline T min(T a, T b)		{	return (a > b ? b : a);	}

	static inline T aton(const char *s);

	T value;
};


template <>	inline float	CMath<float>::abs(float a)	{	return m_fabsf(a);	}
template <>	inline double	CMath<double>::abs(double a)	{	return ::fabs(a);	}
template <>	inline int	CMath<int>::abs(int a)		{	return ::abs(a);	}

template <>	inline float	CMath<float>::floor(float a)	{	return m_floorf(a);	}
template <>	inline double	CMath<double>::floor(double a)	{	return ::floor(a);	}
template <>	inline int	CMath<int>::floor(int a)	{	return a;		}

template <>	inline float	CMath<float>::ceil(float a)	{	return m_ceilf(a);	}
template <>	inline double	CMath<double>::ceil(double a)	{	return ::ceil(a);	}
template <>	inline int	CMath<int>::ceil(int a)	{	return a;		}

template <>	inline float	CMath<float>::sqrt(float a)	{	return m_sqrtf(a);	}
template <>	inline double	CMath<double>::sqrt(double a)	{	return ::sqrt(a);	}
template <>	inline int	CMath<int>::sqrt(int a)	{	return m_sqrtl((unsigned long)a);	}

template <>	inline float	CMath<float>::round(float a)	{	return m_roundf(a);	}
template <>	inline double	CMath<double>::round(double a)	{	return m_round(a);	}
template <>	inline int	CMath<int>::round(int a)	{	return a;	}

template <>	inline float	CMath<float>::aton(const char *s)	{	return ::atof(s);	}
template <>	inline double	CMath<double>::aton(const char *s)	{	return ::atof(s);	}
template <>	inline int	CMath<int>::aton(const char *s)	{	return ::atoi(s);	}


template <typename T>
std::ostream&
operator<<(std::ostream& os, const CMath<T> &lm)
{
	return os << lm.value;
}


#endif
